/****************************************************************************

 This file is part of the QGLViewer library
 Copyright (C) 2002, 2003, 2004  Gilles Debunne (Gilles.Debunne@imag.fr)
 Version 1.3.9-2, released on December 3, 2004.

 http://artis.imag.fr/Members/Gilles.Debunne/QGLViewer

 libQGLViewer is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 libQGLViewer is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with libQGLViewer; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/

#ifndef QGLVIEWER_KEY_FRAME_INTERPOLATOR_H
#define QGLVIEWER_KEY_FRAME_INTERPOLATOR_H

#include <qobject.h>
#include <qtimer.h>

#include "quaternion.h"

// If you compiler complains about incomplete type, uncomment the next line
// #include "frame.h"
// and comment "class Frame;" a few lines below


namespace qglviewer {
  class Camera;
  class Frame;
  /*! \brief A key frame interpolator.
   
  KeyFrameInterpolator implements a classical interpolation of key Frames, using a smooth
  Catmull-Rom interpolation. See the <a href="../examples/keyFrames.html">keyFrames example</a> for
  an illustration of the class.

  A KeyFrameInterpolator holds some key frames, that define a path, and is given a pointer to a
  Frame of your application. When the user startInterpolation(), the KeyFrameInterpolator will make
  this frame() follow the path, updating its position and orientation over time. Here is a typical
  utilization example:
  
  \code
  // The KeyFrameInterpolator kfi is given the Frame that it will drive over time. 
  KeyFrameInterpolator kfi( new Frame() );

  init()
  {
    kfi.addKeyFrame( Frame( Vec(1,0,2), Quaternion() ) );
    kfi.addKeyFrame( new Frame( Vec(2,1,0), Quaternion() ) );
    // ...and so on for all the Key Frames.

    // Ask for a display update after each update of the KeyFrameInterpolator
    connect(&kfi, SIGNAL(interpolated()), SLOT(updateGL()));
    
    kfi.startInterpolation();
  }

  draw()
  {
    kfi.drawPath(); // optional
    glPushMatrix();
    // Retrieve kfi associated frame with kfi.frame(), and use its matrix().
    glMultMatrixd( kfi.frame()->matrix() );
      // Draw your object here. Its position and orientation will be interpolated.
    glPopMatrix();
  }
  \endcode
  
  The keyFrames are defined by a Frame and a time, expressed in seconds. The Frame can be provided
  as a const reference or as a pointer to a Frame (see the addKeyFrame() methods). In the latter
  case, the path will automatically be updated when the Frame is modified (see the Frame::modified()
  signal).

  The time has to be monotonously increasing over keyFrames. When interpolationSpeed() equals 1.0
  (default value), these times correspond to actual user' seconds during the interpolation (provided
  that your QGLViewer::draw() function is fast enough).

  During the interpolation, the KeyFrameInterpolator emits an interpolated() signal, which will
  usually be connected to the viewer's \c updateGL() slot. The finished() signal is emitted at the
  end of an interpolation().
  
  Note that a Camera has 12 potential KeyFrameInterpolator binded to the F1-F12 keys, that can be
  used to drive the Camera along a path, or to restore a saved position (path made of a single key
  frame). Press Alt+F\e x to define a new keyFrame for path \e x. Pressing F\e x plays/pause path
  interpolation. See the <a href="../keyboard.html">keyboard page</a> for details.

  The interpolation can loop (setLoopInterpolation()) and the keyFrame path can be closed
  (setClosedPath()) (not yet implemented). The current interpolationTime() and the
  interpolationSpeed() can also be tuned.

  <h3>Interpolation details</h3>
  
  When the user startInterpolation(), a timer is started, which will update the frame()'s position
  and orientation every interpolationPeriod() milliseconds. This update increases the
  interpolationTime() by interpolationPeriod() * interpolationSpeed() milliseconds.

  If the application main loop is fast enough, and interpolationSpeed() equals 1.0 (default), the
  interpolation will be real-time : the key frames will be reached at their precise keyFrameTime().

  Note that this mechanism ensures that the number of interpolation steps is constant and equal to
  the total path duration() divided by the interpolationPeriod()*interpolationSpeed(). This is
  especially useful for benchmarking or movie creation. The number of frame() updates is hence
  constant, even when real-time interpolation is not possible (for instance because drawing is too
  slow) and the timer frequency cannot be satisfied.
  
  The interpolation is stopped when interpolationTime() is greater than the lastTime()
  (unless loopInterpolation() is \c true) and the finished() signal is then emitted.

  \attention If a Constraint is attached to the frame() (see Frame::setConstraint()), it should be
  desactived before interpolationIsStarted(), otherwise the interpolated motion (computed as if
  there was no constraint) will probably be erroneous.

  <h3>Retrieving interpolated values</h3>

  This code defines a KeyFrameInterpolator, and displays the positions that will be followed by the
  frame() along the path.

  \code
  KeyFrameInterpolator kfi( new Frame() );
  // calls to kfi.addKeyFrame() to define the path.

  float deltaTime = 0.04; // output a position every deltaTime seconds
  for (float time=kfi.firstTime(); time<=kfi.lastTime(); time += deltaTime)
  {
    kfi.interpolateAtTime(time);
    cout << "t=" << time << "\tpos=" << kfi.frame()->position() << endl;
  }
  \endcode

  You may want to temporally disconnect the \c kfi interpolated() signal from the viewers'
  updateGL() slot before calling this code.
  \nosubgrouping */
  class QGLVIEWER_EXPORT KeyFrameInterpolator : public QObject
  {
    // todo closedPath, insertKeyFrames, deleteKeyFrame, replaceKeyFrame
    Q_OBJECT
    
  public:
    KeyFrameInterpolator(Frame* fr=NULL);
    virtual ~KeyFrameInterpolator();

  signals:
    /*! This signal is emitted whenever the frame() state is interpolated. Connect this signal to
    any object that should be notified.

    The emission of this signal triggers the synchonous emission of the frame()
    Frame::interpolated() signal, which may also be useful.

    This signal should especially be connected to your viewer \c updateGL() slot, so that the display
    is updated after every update of the KeyFrameInterpolator frame(). Use code like this :
    \code
    connect(myKeyFrameInterpolator, SIGNAL(interpolated()), SLOT(updateGL()));
    \endcode

    Note that every QGLViewer::camera() holds Camera::nbPaths() KeyFrameInterpolator and that their
    interpolated() signals are automatically connected to the viewer's \c updateGL() slot.

    When several viewers display the same scene, it may be interesting to update \e all the viewers
    when a KeyFrameInterpolator (and probably hence an object) is interpolated. Simply use this code
    to achieve this behavior:
    \code
    QGLViewer::connectSignalToAllViewers(myKeyFrameInterpolator, SIGNAL(interpolated()));
    \endcode

    See also the ManipulatedFrame::manipulated() and SpinningFrame::spinned() signals documentations. */
    void interpolated();

    /*! This signal is emitted when the interpolation reaches the last key frame. Use
    startInterpolation() to start the interpolation and you will be warned with this signal when the
    interpolation is finished.

    \note A call to stopInterpolation() will \e not emit this signal. If loopInterpolation() is
    activated, the signal is emitted each time the interpolation loops back to first keyFrame.

    \note If the interpolationSpeed() is negative, finished() is emitted when interpolationTime()
    reaches firstTime() instead of lastTime(). */
    void finished();

    /*! @name Adding Key Frames */
    //@{
  public slots:
    void addKeyFrame(const Frame& fr);
    void addKeyFrame(const Frame& fr, float time);

    void addKeyFrame(const Frame* fr);
    void addKeyFrame(const Frame* fr, float time);

    void deletePath();
    //@}

    /*! @name Associated Frame */
    //@{
  public:
    /*! The frame that is associated and modified by the KeyFrameInterpolator. When the
    interpolationIsStarted(), this Frame position and orientation will regularly be updated by a
    timer, so that they follow the KeyFrameInterpolator path. Set using setFrame() or with the
    KeyFrameInterpolator() constructor. */
    Frame* frame() const { return frame_; };

  public slots:
    void setFrame(Frame* fr);
    //@}

    /*! @name Path parameters */
    //@{
  public:
    /*! Number of keyFrames used in the interpolation. Use addKeyFrame() to add new keyFrames. */
    int numberOfKeyFrames() const { return keyFrame_.count(); };
    
    Frame keyFrame(int i) const;
    Vec keyFramePosition(int i) const;
    Quaternion keyFrameOrientation(int i) const;
    float keyFrameTime(int i) const;
    float duration() const;
    float firstTime() const;
    float lastTime() const;
    //@}

    /*! @name Interpolation parameters */
    //@{
  public:
    /*! Current interpolation time (in seconds) along the KeyFrameInterpolator path, updated when
    interpolationIsStarted(). Can be set directly with setInterpolationTime() or
    interpolateAtTime(). */
    float interpolationTime() const { return interpolationTime_; };
    /*! Returns the current interpolation speed. Default value is 1.0, which means that the times
    associated with the keyFrames (see keyFrameTime() and addKeyFrame()), expressed in seconds, will
    be matched during the interpolation, provided that your display is fast enough.
      
    A negative value will result in a reverse interpolation of the keyFrames. See also
    interpolationPeriod(). */
    float interpolationSpeed() const { return interpolationSpeed_; };
    /*! Returns the current interpolation period, expressed in milliseconds. The update of the
    frame() state will be done by a timer at this period when interpolationIsStarted().

    This period (multiplied by interpolationSpeed()) is also added to the interpolationTime() at
    each update, and the frame() state is modified accordingly (see interpolateAtTime()). Default
    value is 40 milliseconds. */
    int interpolationPeriod() const { return period_; };
    /*! Whether or not (default) the interpolation will be played in (infinite) loop when
    startInterpolation() is called (until stopInterpolation() is called). When \p false, the
    interpolation will stop when interpolationTime() reaches lastTime().

    Use setClosedPath() to create a closed path with the keyFrames. */
    bool loopInterpolation() const { return loopInterpolation_; };
    /*! Whether or not (default) the path defined by the keyFrames is a closed loop. When \p true,
    the last and the first KeyFrame are linked by a new spline segment. 
      
    Use setLoopInterpolation() to create a continuous animation over the entire path.
    \attention The closed path feature is not yet implemented. */
    bool closedPath() const { return closedPath_; };
  public slots:
    /*! Sets the current interpolation time. See also interpolationTime().

    \attention The frame() state is not affected by this method. Use this function to define the
    starting time of a future interpolation (see startInterpolation()). Use interpolateAtTime() to
    actually interpolate at a given time. */
    void setInterpolationTime(float time) { interpolationTime_ = time; };
    /*! Sets the interpolation speed. Negative values are possible. See interpolationSpeed(). */
    void setInterpolationSpeed(float speed) { interpolationSpeed_ = speed; };
    /*! Sets the interpolationPeriod(). */
    void setInterpolationPeriod(int msecs) { period_ = msecs; };
    /*! Sets the loopInterpolation() value. */
    void setLoopInterpolation(bool loop=true) { loopInterpolation_ = loop; };
    /*! Sets the closedPath() value. \attention The closed path feature is not yet implemented. */
    void setClosedPath(bool closed=true) { closedPath_ = closed; };
    //@}

    
    /*! @name Interpolation */
    //@{
  public:
    /*! Returns \p true when the interpolation is being performed. Calls to startInterpolation(),
    stopInterpolation() or toggleInterpolation() migth modify this state. */
    bool interpolationIsStarted() const { return interpolationStarted_; };
  public slots:
    virtual void interpolateAtTime(float time);
    void startInterpolation(int msecs = -1);
    void stopInterpolation();
    void resetInterpolation();
    /*! Calls startInterpolation() or stopInterpolation(), depending on interpolationIsStarted(). */
    void toggleInterpolation() { if (interpolationIsStarted()) stopInterpolation(); else startInterpolation(); };
    //@}
    
    /*! @name Path drawing */
    //@{
  public:
    virtual void drawPath(int mask=1, int nbFrames=6, float scale=1.0f);
    virtual void drawPathModifyGLState(int mask, int nbFrames, float scale);
    //@}

  public:
    /*! @name XML representation */
    //@{
    virtual QDomElement domElement(const QString& name, QDomDocument& doc) const;
    virtual void initFromDOMElement(const QDomElement& e);
    //@}
      
  private slots:
    virtual void update();
    virtual void invalidateValues() { valuesAreValid_ = false; pathIsValid_ = false; splineCacheIsValid_ = false; };
    
  private:
    // Copy constructor and opertor= are declared private and undefined
    // Prevents everyone from trying to use them
    KeyFrameInterpolator(const KeyFrameInterpolator& kfi);
    KeyFrameInterpolator& operator=(const KeyFrameInterpolator& kfi);

    void updateCurrentKeyFrameForTime(float time);
    void updateModifiedFrameValues();
    void updateSplineCache();

    // Internal private KeyFrame representation
    class KeyFrame
    {
    public:
      KeyFrame(const Frame& fr, float t);
      KeyFrame(const Frame* fr, float t);

      Vec position() const { return p_; }
      Quaternion orientation() const { return q_; }
      Vec tgP() const { return tgP_; }
      Quaternion tgQ() const { return tgQ_; }
      float time() const { return time_; }
      const Frame* frame() const { return frame_; }
      void updateValuesFromPointer();
      void flipOrientation(const Quaternion& prev);
      void computeTangent(const KeyFrame* const prev, const KeyFrame* const next);
    private:
      Vec p_, tgP_;
      Quaternion q_, tgQ_;
      float time_;
      const Frame* const frame_;
    };

    // K e y F r a m e s
    mutable QPtrList<KeyFrame> keyFrame_;
    // 4 succesive frames. interpolationTime_ is between index 1 and 2.
    QPtrListIterator<KeyFrame>* currentFrame_[4];
    // If you compiler complains about incomplete type, include frame.h (see top of this file)
    QValueVector<Frame> path_;
    
    // A s s o c i a t e d   f r a m e
    Frame* frame_;
    
    // R h y t h m
    QTimer timer_;
    int period_;
    float interpolationTime_;
    float interpolationSpeed_;
    bool interpolationStarted_;

    // M i s c
    bool closedPath_;
    bool loopInterpolation_;

    // C a c h e  v a l u e s   a n d   f l a g s
    bool pathIsValid_;
    bool valuesAreValid_;
    bool currentFrameValid_;
    bool splineCacheIsValid_;
    Vec v1, v2;
  };
  
} // namespace qglviewer

#endif // QGLVIEWER_KEY_FRAME_INTERPOLATOR_H
